#include "polygon.h"
#include "randomizer.h" 
#include "gl\gl.h"
#include "gl\glu.h"
#include <assert.h>

//#define QUERYLINEWIDTHRANGE
/* If your lines aren't  showing up thick it may be that your particular OpenGL hardware
doesn't support thick lines.  You can find out by commenting in QUERYLINEWIDTHRANGE
and setting a break point after the corresponding block of code in cPolygon::draw.
The code is odd-looking code, but is a standard way to query OpenGL about what it can do.*/

//******** CONST STATICS **********************************************
const int cPolygon::MF_RADIUS =		0x00000008;
const int cPolygon::MF_ROTATION =	0x00000010;
const int cPolygon::MF_COLOR =		0x00000020;
const int cPolygon::MF_FILLING =	0x00000040;
const int cPolygon::MF_EDGED =		0x00000080;
const int cPolygon::MF_LINEWIDTH =	0x00000100;
const int cPolygon::MF_VERTCOUNT =	0x00000200;
const int cPolygon::MF_ALL = cPolygon::MF_RADIUS | cPolygon::MF_ROTATION |
	cPolygon::MF_COLOR | cPolygon::MF_FILLING |
	cPolygon::MF_LINEWIDTH |  	cPolygon::MF_VERTCOUNT;
Real cPolygon::MINRADIUS =  0.3;
Real cPolygon::MAXRADIUS = 1.0;
Real cPolygon::PRISMDZ = 0.2;
const int cPolygon::DF_WIREFRAME = 1;

//Helper class---------------------------

class cIntPair
{
public:
	int x, y;
	cIntPair(int ix=0, int iy=0):x(ix),y(iy){}
};

//------------Constructors, Destructor, and Copiers---------------
void cPolygon::_initializer() //used by the constructors
{
	//The two CArrays are set to size 0 by their default constructors
	//Three helper fields
	_pcolorstyle = new cColorStyle();
	_closed = FALSE;
	//Former cSprite fields
	_prismdz = cPolygon::PRISMDZ;
	setCenter(cVector(0.0, 0.0, 0.0)); //Sets last column of _attitude.
	_radius = 1.0;
	//Default constructor of _spriteattidute will be an identity matrix.
	_tumbleaxis = cSpin(cRandomizer::pinst()->randomReal(5*PI/180.0, 15*PI/180.0),
		cVector::randomUnitVector()); //cSpin constuctor takes spinangle, spinaxis args.
	_tumble = cMatrix::rotation(_tumbleaxis);
	cSpin spin = cSpin(cRandomizer::pinst()->randomReal(0.5*PI/180.0, 0.15*PI/180.0),
		cVector::randomUnitVector()); //cSpin constuctor takes spinangle, spinaxis args.
	_precess = cMatrix::rotation(spin);
}

// The default constructor makes an empty polygon with standard	decorations.
cPolygon::cPolygon()
{
	_initializer();
} 

cPolygon::cPolygon(int n, cVector *pverts, cColorStyle *pcolorstyle)
{
	_initializer();
	_vectorvert.resize(n);
	for(int i=0; i<n; i++)
		_vectorvert[i] = pverts[i];
	if (pcolorstyle)
		setColorStyle(new cColorStyle(pcolorstyle)); //Otherwise keep the default.
	_closed = TRUE;
	fixCenterAndRadius();
}

cPolygon::cPolygon(int vertcount)
{
	_initializer();
	setRegularPolygon(vertcount, cVector(0.0, 0.0), 1.0, 0.0);
}

cPolygon::~cPolygon()
{
	delete _pcolorstyle;
}

void cPolygon::setColorStyle(cColorStyle *pcolorstyle)
{
	delete _pcolorstyle;
	_pcolorstyle = pcolorstyle;
}

//------------------Copy and operator= -----------------
void cPolygon::copy(const cPolygon &cpoly)
{
	if (this == &cpoly)
		return;
/* Because our class has some vector fields, we can't use the default overloaded
copy constructor and operator=.  So as to avoid having to maintain similar code
for these two different methods, we write a helper copy function that both
the copy constructor and the operator= can use. */
	setSize(cpoly.vertCount());
	for (int i=0; i<vertCount(); i++)
		setVertex(i, cpoly.getVertex(i));
	//Decoration fields
	setColorStyle(new cColorStyle(cpoly._pcolorstyle));
	//Extra fields
	_closed = cpoly._closed;
	_radius = cpoly._radius; 
	_attitude = cpoly._attitude; 
	_prismdz = cpoly._prismdz; 
	_tumbleaxis = cpoly._tumbleaxis;
	_tumble = cpoly._tumble;
	_precess = cpoly._precess;
}

cPolygon cPolygon::operator=(const cPolygon &cpoly)
{
	if (this == &cpoly)
		return *this;
	copy(cpoly);
	return *this;
}

cPolygon::cPolygon(const cPolygon &cpolygon)
{
	_initializer();
	copy(cpolygon);
}

cPolygon::cPolygon(cPolygon *ppolygon)
{
	_initializer();
	copy(*ppolygon);
}

//-----------------Geometric mutators-----------------

void cPolygon::fixCenterAndRadius()
{
//Calculate the centroid
	/* We plan to divide by vertCount, so don't allow it to be zero. */
	if (!vertCount())
		return;
	cVector centroid; //Default constuctor starts at zeroVector.
	for (int i=0; i<vertCount(); i++)
		centroid += _vectorvert[i];
	centroid /= vertCount(); // We already made sure this divisor isn't zero.
//Move the centroid to the origin
	for (i=0; i<vertCount(); i++)
		_vectorvert[i] -= centroid;
//Fix Radius as furthest vertex from the origin.
	Real distance = 0.0; //Start with this and look for the biggest one.
	Real testdistance;
	for ( i=0; i<vertCount(); i++)
	{
		testdistance = _vectorvert[i].magnitude(); 
		if (testdistance > distance)
			distance = testdistance;
	}
	_radius = distance; //_radius is a raw number.
}

void cPolygon::setRadius(Real newradius)
{
	fixCenterAndRadius(); /* In case you haven't done this yet, for instance if you've built
		the polygon with a bunch of setVertex or addVertex calls and you are now calling setRadius. */
	if (!newradius)
		return; //Don't allow zero radius.
	assert(radius()); //else it's hopeless
	Real scalefactor = newradius/radius();
	for (int i=0; i<vertCount(); i++)
		_vectorvert[i] *= scalefactor; // Centroid is origin so we can scale around it.
	_radius = scalefactor * _radius; /* _radius measures the newly scaled geometry of the
		polygon around the origin.  Note  newradius = attitude._scalefactor()*_radius
		rather than newradius = _radius, although quite often the scalefactor is in fact 1.0. */
}

vector<cVector>* cPolygon::transformedVert(const cMatrix &M)
{
	_transformedvert.resize(vertCount());
	for (int i=0; i<vertCount(); i++)
		_transformedvert[i] = M * _vectorvert[i];
	return &_transformedvert;
}

//----------Geometric accessor-------------
cVector cPolygon::getVertex(int n) const
{
	return _vectorvert.at(n);
}

Real cPolygon::angle()
{
	if (!vertCount())
		return 0.0;
	return angleBetween(_vectorvert[0]-center(), cVector(1.0, 0.0, 0.0));
}
//----------------Mutators----------------------

void cPolygon::setSize(int newcount)
{ 
	_vectorvert.resize(newcount);
}

void cPolygon::addVertex(const cVector &vect)
{
	setSize(vertCount() + 1);
	_vectorvert[vertCount()-1] = vect;
	_closed = FALSE;
}

void cPolygon::setVertex(int n, const cVector &vect)
{
	if (vertCount() < n+1)
		setSize(n+1);
	_vectorvert[n] = vect;
}

void cPolygon::setRegularPolygon(int vertexcount, cVector center, Real newradius,
	Real initangle)
{
	Real angle = initangle;

	setSize(vertexcount);
	for (int i=0; i<vertexcount; i++)
	{
		_vectorvert[i] = center + newradius*cVector(cos(angle), sin(angle));
		angle += (2*PI)/vertexcount;
	}
	_closed = TRUE;
	setCenter(center);
	_radius = newradius;
	setRadius(_radius);
}

void cPolygon::changeVertexcount(int updown)
{
	int newvertcount = _vectorvert.size() + updown;
	if (newvertcount < 2)
		newvertcount = 2;
	setRegularPolygon(newvertcount, center(), radius(), angle());
}

void cPolygon::setStarPolygon(int vertexcount, int step, cVector center,
	Real newradius, Real initangle)
{
	Real angle;
	int i = 0, stepi = 0;

	if (step < 1)
		step = 1; //Avoid endless loop you'd get if step == 0.
	setSize(vertexcount);
	while (!i || stepi)
	/* Do the while loop for the first step (when !i) and for any
	succeding step where stepi is not yet divisible by vertexcount*/
	{
		angle = initangle + (stepi * 2.0 * PI)/vertexcount;
		_vectorvert[i] = center + newradius*cVector(cos(angle), sin(angle));
		i++;
		stepi = (i*step)%vertexcount;
	}
	setSize(i); //There are i actual vertexes that got used.
	_closed = TRUE;
	setCenter(center);
	setRadius(_radius);
}

void cPolygon::randomize(int mutationflags)
{
	if (mutationflags & cPolygon::MF_RADIUS)
		setRadius(cRandomizer::pinst()->randomReal(
			cPolygon::MINRADIUS, cPolygon::MAXRADIUS)); 
	if (mutationflags & MF_ROTATION)
		rotate(cSpin(cRandomizer::pinst()->randomReal(-PI, PI),	cVector::randomUnitVector())); 
	//	rotate(cRandomizer::pinst()->randomReal(0.0, 2.0*PI));
	_pcolorstyle->randomize(mutationflags);
//Number of vertices
	if (mutationflags & cPolygon::MF_VERTCOUNT)
	{
		int newcount;
		if  (!vertCount()) 
		{//Creating a new polygon 
			if (cRandomizer::pinst()->randomBool(0.5))
				setRegularPolygon(cRandomizer::pinst()->random(2,9), center(), _radius, angle());
			else
			{
				newcount = cRandomizer::pinst()->random(5, 14);
				setRandomStarPolygon(newcount, newcount);
			}
		}
		else
		{
			if (cRandomizer::pinst()->randomBool())
			{
				newcount = cRandomizer::pinst()->random(3, 10);
				setRegularPolygon(newcount,	center(), _radius, angle());
			}
			else
			{
				newcount = cRandomizer::pinst()->random(5, 22);
				setRandomStarPolygon(newcount, newcount);
			}
		}
	}
//Fixups
	setRadius(_radius); 
}

//-----------------Randomizing Functions---------------
int GCD(int a, int b)
{ // Euclid's algorithm for the greatest common divisor of a and b.
	if (a < b)
	{
		int temp = a;
		a = b;
		b = temp;
	}
	if (!b)
		return a;
	int q = a/b;
	int r = a - b*q;
	return GCD(q,r);
}

cIntPair next_rel_prime_pair(const cIntPair &startpoint)
{
/* We use this to find pairs of integers to use for the (count, step) arguments
of setStarPolygon that we draw by starting with count points and skipping step
of them as we move from each vertex to the next.  You only hit all count vertices
if step and count are relatively prime, that is, if they have a greatest common
divisor or GCD of 1.  The traditional pentagram, for instance, uses (5,2). This
function takes a (count, step) pair and finds the next "good" pair that's relatively
prime.  Our notion of "good" has to do with the following  considerations.
First of all, we don't want to bother with a step of 1, as this gives a normal,
non-star polygon.  Second, step and count-step will give the same star, so we 
only look at step values which are <= n/2. Third, when you have a largish count and 
small step like 2, the star looks like a hollow round doily, which is pretty
nice, but we could hange the code to rule it out. Fourth, if the step is close to
n/2 we get something like an asterisk, which I think is okay too, but this would be
another thing that we could change the code to avoid. */
	int newcount = startpoint.x;
	int newstep = startpoint.y;
	
	if (newstep > newcount/2)
		newstep = newcount - newstep;

	while (TRUE) //Stay in here until a return bails you out.
	{
		newstep++;
		if (newstep >= newcount/2) /* Avoid repeating the cases.  If you don't
			like asterisks use something like newcount/3 here instead. */
		{
			newcount++;
			newstep = 2; /* Allow doily, though you could put 3 here, or even
				__max(2, newcount/4); */
		}
		if (GCD(newcount, newstep) == 1)
			return cIntPair(newcount, newstep);
	}
}

void cPolygon::setRandomRegularPolygon(int mincount, int maxcount)
{
	setRegularPolygon(cRandomizer::pinst()->random(mincount, maxcount), center(),
		_radius, angle());
}

void cPolygon::setRandomStarPolygon(int mincount,
	int maxcount)
{
	cIntPair starpoint(cRandomizer::pinst()->random(mincount, maxcount),
		cRandomizer::pinst()->random(2,maxcount/2));

	if (GCD(starpoint.x, starpoint.y) != 1)
		starpoint = next_rel_prime_pair(starpoint);
	setStarPolygon(starpoint.x, starpoint.y, center(),
		_radius, angle());
}

void cPolygon::setRandomAsteroidPolygon(int mincount, int maxcount, Real spikiness)
{			
	Real angle = 0.0, anglestep;
	int vertexcount = cRandomizer::pinst()->random(mincount, maxcount);
	setSize(vertexcount);
	anglestep = 2*PI/vertexcount;
	for (int i=0; i<vertexcount && angle < 2.0*PI; i++)
	{
		cVector temp = 
		_vectorvert[i] = cRandomizer::pinst()->randomReal(1.0 - spikiness, 1.0 + spikiness) *
			cVector(cos(angle), sin(angle));
		angle += cRandomizer::pinst()->randomReal(0.0, 2.0*anglestep);
	}
	setSize(i);
	setClosed(TRUE);
	fixCenterAndRadius();
	setRadius(_radius);
}

void cPolygon::draw(int drawflags)
{
//Doesn't really work right for star polygons. 
	::glMatrixMode(GL_MODELVIEW);
	::glPushMatrix();
	::glMultMatrixf(_attitude.transpose().elements());
	/* Get a normal ready in case you are using lighting.  Assuming
		the polygon vertices are labelled CCW.   For our standard polys, upnormal 
		and downnormal will	in fact just be the pos and neg ZAXIS.   */
	cVector upnormal;
	if(vertCount() >= 3)
		upnormal = (getVertex(1) - getVertex(0)) *
		(getVertex(2) - getVertex(0));
	if (upnormal.isZero())
		upnormal = cVector::ZAXIS;
	upnormal.normalize();
	if (filled())
	{
		setVertexColor(fillColor(), 0.9); //For use with no lighting
		setMaterialColor(fillColor()); //For use with lighting
		::glNormal3f(upnormal.x(), upnormal.y(), upnormal.z()); //For lighting.
		::glBegin(GL_TRIANGLE_FAN); 
			::glVertex3d(0.0, 0.0, 0.0); /* Draw as if center is at origin,
				and let the _attitude take care of moving it. */
			for (int i=0; i< this->vertCount(); i++)
				getVertex(i).glVertex();
			getVertex(0).glVertex();
		::glEnd();
	}
	if (edged()) 
	{
		setVertexColor(lineColor()); //For use with no lighting
		setMaterialColor(lineColor()); //For use with lighting
		::glEnable(GL_LINE_SMOOTH);
		::glLineWidth(lineWidth()); //Must set this outside a glBegin/glEnd block.
#ifdef QUERYLINEWIDTHRANGE
		float params[2];
		glGetFloatv(GL_LINE_WIDTH_RANGE, params);
		int min = params[0];
		int max = params[1];
#endif //QUERYLINEWIDTHRANGE
		::glBegin(GL_LINE_LOOP); 
			for (int i=0; i<vertCount(); i++)
				getVertex(i).glVertex();
		::glEnd();
	}
	::glPopMatrix();
}


